
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <fcntl.h>

#include "alloc-ioctls.h"

static int verbose_mode = 0;
static int null_mode = 0;
static char *root_dir = 0;

static int this_block_size = 0;
static int this_total_blocks = 0;

/*
Given the size of a file and the filesystem block size,
return how many blocks are actually consumed on disk,
taking into acount indirect blocks.
*/

long long blocks_from_size( long long size, long long blocksize )
{
	long long direct_block_max = 12;
	long long branching_factor = this_block_size/4;
	long long indirect_blocks        = 0;
	long long doubly_indirect_blocks = 0;
	long long triply_indirect_blocks = 0;

	long long datablocks = size/blocksize;

	if(size%blocksize) datablocks++;

	if(datablocks>direct_block_max) {
		indirect_blocks = 1 + (datablocks-direct_block_max)/branching_factor;
		if(indirect_blocks>1) {
			doubly_indirect_blocks = 1+(indirect_blocks-1)/branching_factor;
			if(doubly_indirect_blocks>1) {
				triply_indirect_blocks = 1;
			}
		}
	}
	

	return datablocks + indirect_blocks + doubly_indirect_blocks + triply_indirect_blocks;
}

int fixalloc( const char *path, int allocparent )
{
	DIR *dir;
	struct dirent *d;
	int total = 0;
	char fullpath[4096];
	int fd, result;
	struct stat info;
	int state[3];

	if(verbose_mode) printf("%s:\n",path);

	fd = open(path,O_RDONLY);
	if(fd<0) goto badaccess;

	fstat(fd,&info);

	if(info.st_ino==allocparent) {
		state[0] = this_total_blocks;
		state[1] = 0;
		state[2] = allocparent;		
		if(!null_mode) {
			result = ioctl(fd,EXT2_IOC_SETALLOC,state);
			if(result<0) goto badalloc;
		}
	} else {
		result = ioctl(fd,EXT2_IOC_LSALLOC,state);
		if(result<0) goto badalloc;

		if(state[0]!=0) {
			allocparent = info.st_ino;
		} else {
			if(state[2]!=allocparent) {
				printf("WARNING: %s had incorrect parent inode %d, changing to %d\n",path,state[3],allocparent);
				state[2] = allocparent;
				if(!null_mode) {
					result = ioctl(fd,EXT2_IOC_SETALLOC,state);
					if(result<0) goto badalloc;
				}
			}
		}
	}

	close(fd);

	dir = opendir(path);
	if(!dir) goto badaccess;

	while((d=readdir(dir))) {
		if(!strcmp(d->d_name,".")) continue;
		if(!strcmp(d->d_name,"..")) continue;

		sprintf(fullpath,"%s/%s",path,d->d_name);

		result = lstat(fullpath,&info);
		if(result<0) goto badaccess;

		result = blocks_from_size(info.st_size,this_block_size);
		total += result;

		if(S_ISDIR(info.st_mode)) {
			total += fixalloc(fullpath,allocparent);
		} else {
			if(verbose_mode) printf("%s: %d blocks used\n",path,result);
		}

	}

	closedir(dir);

	fd = open(path,O_RDONLY);
	if(fd<0) goto badaccess;

	result = ioctl(fd,EXT2_IOC_LSALLOC,state);
	if(result<0) goto badalloc;

	if(state[0]!=0) {
		state[1] = total;
		total = state[0];
		if(!null_mode) {
			result = ioctl(fd,EXT2_IOC_SETALLOC,state);
			if(result<0) goto badalloc;
		}
		if(state[1]>state[0]) printf("WARNING: %s EXCEEDS ALLOCATION \n",path);
		if(verbose_mode) printf("%s: %d blocks used of %d blocks allocated\n",path,state[1],state[0]);
	} else {
		if(verbose_mode) printf("%s: %d blocks used\n",path,total);
	}

	close(fd);

	return total;

	badalloc:
	if(errno==ENOTTY) {
		fprintf(stderr,"ERROR: This filesystem does not support allocations.\n");
	} else {
		fprintf(stderr,"ERROR: Couldn't access allocation in %s: %s\n",path,strerror(errno));
	}
	exit(1);

	badaccess:
	fprintf(stderr,"ERROR: Couldn't access %s: %s\n",path,strerror(errno));
	exit(1);

}

static void show_use()
{
	fprintf(stderr,"Use: fixalloc [options] <path>\n");
	fprintf(stderr,"Where options are:\n");
	fprintf(stderr,"  -v         Verbose mode; show all activity.\n");
	fprintf(stderr,"  -n         Scan filesystem, but do not make changes.\n");
	fprintf(stderr,"  -s <size>  Force this top-level allocation size.  (default is 90%% of blocks.)\n");
	fprintf(stderr,"  -h         Show these options.\n");
}

int main( int argc, char *argv[] )
{
	struct stat info;
	struct statfs fsinfo;
	char c;

        while((c=getopt(argc, argv, "vns:h"))!=(char)-1) {
		switch(c) {
			case 'v':
				verbose_mode++;
				break;
			case 'n':
				null_mode=1;
				break;
			case 's':
				this_total_blocks = atoi(optarg);
				break;
			default:
				fprintf(stderr,"fixalloc: unknown option -%c (try -h for help)\n",optopt);
				return 1;
			case 'h':
				show_use();
				return 1;
		}
	}

	if(optind>=argc) {
		fprintf(stderr,"fixalloc: missing argument! (try -h for help)\n");
		return 1;
	}

	root_dir = argv[optind];

	if(stat(root_dir,&info)!=0 || statfs(root_dir,&fsinfo)!=0 ) {
		fprintf(stderr,"fixalloc: %s: %s\n",root_dir,strerror(errno));
		return 1;
	}

	this_block_size = fsinfo.f_bsize;
	if(this_total_blocks==0) {
		this_total_blocks = fsinfo.f_blocks*.9;
		if(verbose_mode) printf("%s: root allocation will be %d blocks = 90%% of %d blocks in filesystem\n",root_dir,this_total_blocks,fsinfo.f_blocks);
	} else {
		if(this_total_blocks>fsinfo.f_blocks) {
			fprintf(stderr,"ERROR: root allocation (%d blocks) larger than filesystem (%d blocks)\n",this_total_blocks,fsinfo.f_blocks);
			return 1;
		}
		if(verbose_mode) printf("%s: root allocation forced to %d blocks\n",root_dir,this_total_blocks);
	}

	fixalloc(root_dir,info.st_ino);

	return 0;
}
