
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/statfs.h>

#include "alloc-ioctls.h"

static int recursive_mode = 0;
static int recursive_depth = 0;
static char units = 'M';

void printf_units( int blocks, int block_size )
{
	double scale=1.0;

	switch(units) {
		case 'K':
			scale = 1024.0;
			break;
		case 'M':
			scale = 1024.0*1024;
			break;
		case 'G':
			scale = 1024.0*1024*1024;
			break;
		case 'T':
			scale = 1024.0*1024*1024*1024;
			break;
		default:
			printf("%9d ",blocks);
			return;
			break;
	}

	printf("%8.2lf%c ",block_size/scale*blocks,units);
}

int lsalloc( const char *path, int depth, int depth_to_go )
{
	struct stat info;
	struct statfs fsinfo;
	int status[3];
	int fd;
	int result;

	if(recursive_mode && depth_to_go<=0) return 0;

	fd = open(path,O_RDONLY);
	if(fd<0) {
		fprintf(stderr,"couldn't open %s: %s\n",path,strerror(errno));
		return 1;
	}

	fstat(fd,&info);

	if(S_ISDIR(info.st_mode)) {
		result = ioctl(fd,EXT2_IOC_LSALLOC,status);
		if(result<0) {
			fprintf(stderr,"couldn't get allocation: %s\n",
				errno==ENOTTY ?
				"filesystem doesn't support allocations" :
				strerror(errno));
			return 1;
		}

		if(status[0]==0) {
			if(depth==0) {
				printf("     -      -   -%% %s\n",path);
			}
		} else {
			fstatfs(fd,&fsinfo);
			printf_units(status[1],fsinfo.f_bsize);
			printf_units(status[0],fsinfo.f_bsize);
			printf("%3d%% %s\n",(int)(100.0*status[1]/status[0]),path);
			depth++;
			depth_to_go--;
		}
	}

	close(fd);

	if(S_ISDIR(info.st_mode) && recursive_mode) {
		DIR *dir;
		struct dirent *d;

		dir = opendir(path);
		if(!dir) {
			fprintf(stderr,"couldn't opendir %s: %s\n",path,strerror(errno));
			return 1;
		}

		while((d=readdir(dir))) {
			char newpath[PATH_MAX];
			if(!strcmp(d->d_name,".")) continue;
			if(!strcmp(d->d_name,"..")) continue;
			sprintf(newpath,"%s/%s",path,d->d_name);
			lsalloc(newpath,depth,depth_to_go);
		}
		closedir(dir);
	}

	return 0;
}

static void show_use()
{
	fprintf(stderr,"Use: lsalloc [options] <path>\n");
	fprintf(stderr,"Where options are:\n");
	fprintf(stderr,"  -r         Recursively show allocations.\n");
	fprintf(stderr,"  -d <depth> Recursively show allocations to this depth. (e.g. -d 2)\n");
	fprintf(stderr,"  -u <units> Display sizes with these units.  (K, M, G, T, or B for blocks.)\n");
	fprintf(stderr,"  -h         Show these options.\n");
}

int main( int argc, char *argv[] )
{
	char c;
	const char *path;

        while((c=getopt(argc, argv, "d:ru:h"))!=(char)-1) {
		switch(c) {
			case 'd':
				recursive_depth = atoi(optarg);
				recursive_mode = 1;
				break;
			case 'r':
				recursive_depth = 999999;
				recursive_mode = 1;
				break;
			case 'u':
				units = optarg[0];
				break;
			default:
				fprintf(stderr,"lsalloc: unknown option -%c (try -h for help)\n",optopt);
				return 1;
			case 'h':
				show_use();
				return 1;
		}
	}


	if(optind>=argc) {
		path = ".";
	} else {
		path = argv[optind];

	}

	printf("     USED     TOTAL  PCT PATH\n");
	lsalloc(path,0,recursive_depth);

	return 0;
}
