
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/statfs.h>

#include "alloc-ioctls.h"

static int block_size;

unsigned convert_size( const char *str )
{
	int fields;
	char suffix;
	unsigned size;
	unsigned blocks;

	fields = sscanf(str,"%u%c",&size,&suffix);
	if(fields==0) goto badsize;
	if(fields==1) return size;

	switch(toupper(suffix)) {
		case 'B':
			return blocks;
		case 'K':
			return size*1024.0/block_size;
		case 'M':
			return size*1024.0*1024.0/block_size;
		case 'G':
			return size*1024.0*1024.0*1024.0/block_size;
	}

	badsize:
	fprintf(stderr,"mkalloc: invalid size: %s\n",str);
	exit(1);

}

int main( int argc, char *argv[] )
{
	int fd;
	int result;
	struct statfs fsinfo;
	unsigned blocks;
	const char *path;

	if(argc!=3) {
		fprintf(stderr,"Use: mkalloc <dir> <size>\n");
		fprintf(stderr,"Where <dir> may be:\n");
		fprintf(stderr,"   - A new or empty directory to be made into an allocation.\n");
		fprintf(stderr,"   - An existing allocation to modify.\n");
		fprintf(stderr,"Where <size> may be:\n");
		fprintf(stderr,"   - The number of blocks in the allocation\n");
		fprintf(stderr,"   - The total size of the allocation, in K, M, G, or T.\n");
		return 1;
	}

	path = argv[1];

	result = mkdir(path,0777);
	if(result!=0) {
		if(errno==EEXIST) {
			/* no problem, continue */
		} else {
			fprintf(stderr,"mkalloc: couldn't create %s: %s\n",path,strerror(errno));
			return 1;
		}
	}

	fd = open(path,O_RDONLY);
	if(fd<0) {
		fprintf(stderr,"mkalloc: couldn't open %s: %s\n",path,strerror(errno));
		return 1;
	}

	fstatfs(fd,&fsinfo);
	block_size = fsinfo.f_bsize;

	blocks = convert_size(argv[2]);

	result = ioctl(fd,EXT2_IOC_MKALLOC,blocks);
	if(result<0) {
		if(errno==ENOTTY) {
			fprintf(stderr,"mkalloc: This filesystem does not support allocations.\n");
		} else {
			fprintf(stderr,"mkalloc: couldn't make allocation: %s\n",strerror(errno));
		}
		return 1;
	} else {
		printf("mkalloc: %s allocated %d blocks.\n",path,blocks);
		return 0;
	}
}
